@@ -94,6 +94,7 @@ static const char *user_agent = "rspamc";
 static const char *files_list = nullptr;
 static const char *queue_id = nullptr;
 static const char *log_tag = nullptr;
+static const char *output_body = nullptr;
 static std::string settings;
 static std::string neural_train;
 static std::string neural_rule;
@@ -200,6 +201,8 @@ static GOptionEntry entries[] =
 		 "Set Queue-ID header for the request", nullptr},
 		{"log-tag", '\0', 0, G_OPTION_ARG_STRING, &log_tag,
 		 "Set Log-Tag header for the request", nullptr},
+		{"output-body", '\0', 0, G_OPTION_ARG_FILENAME, &output_body,
+		 "Save rewritten message body to file", nullptr},
 		{"settings", '\0', 0, G_OPTION_ARG_CALLBACK, (void *) &rspamc_settings_callback,
 		 "Set Settings header as JSON/UCL for the request", nullptr},
 		{nullptr, 0, 0, G_OPTION_ARG_NONE, nullptr, nullptr, nullptr}};
@@ -1848,9 +1851,6 @@ rspamc_mime_output(FILE *out, ucl_object_t *result, GString *input,
 		fmt::format_to(std::back_inserter(added_headers), "X-Spam-Scan-Time: {:.3}{}",
 					   time, line_end);
 
-		/*
-		 * TODO: add milter_headers support here
-		 */
 		if (is_spam) {
 			fmt::format_to(std::back_inserter(added_headers), "X-Spam: yes{}", line_end);
 		}
@@ -1908,6 +1908,47 @@ rspamc_mime_output(FILE *out, ucl_object_t *result, GString *input,
 			}
 		}
 
+		/* Process milter.add_headers */
+		const auto *milter = ucl_object_lookup(result, "milter");
+		if (milter) {
+			const auto *add_headers = ucl_object_lookup(milter, "add_headers");
+			if (add_headers && ucl_object_type(add_headers) == UCL_OBJECT) {
+				it = nullptr;
+				while ((cur = ucl_object_iterate(add_headers, &it, true)) != nullptr) {
+					const char *header_name = ucl_object_key(cur);
+					if (ucl_object_type(cur) == UCL_OBJECT) {
+						/* Handle {order: N, value: "..."} format */
+						const auto *value_obj = ucl_object_lookup(cur, "value");
+						if (value_obj) {
+							fmt::format_to(std::back_inserter(added_headers), "{}: {}{}",
+										   header_name, ucl_object_tostring(value_obj), line_end);
+						}
+					}
+					else if (ucl_object_type(cur) == UCL_STRING) {
+						fmt::format_to(std::back_inserter(added_headers), "{}: {}{}",
+									   header_name, ucl_object_tostring(cur), line_end);
+					}
+					else if (ucl_object_type(cur) == UCL_ARRAY) {
+						ucl_object_iter_t header_it = nullptr;
+						const ucl_object_t *header_value;
+						while ((header_value = ucl_object_iterate(cur, &header_it, true)) != nullptr) {
+							if (ucl_object_type(header_value) == UCL_OBJECT) {
+								const auto *value_obj = ucl_object_lookup(header_value, "value");
+								if (value_obj) {
+									fmt::format_to(std::back_inserter(added_headers), "{}: {}{}",
+												   header_name, ucl_object_tostring(value_obj), line_end);
+								}
+							}
+							else {
+								fmt::format_to(std::back_inserter(added_headers), "{}: {}{}",
+											   header_name, ucl_object_tostring(header_value), line_end);
+							}
+						}
+					}
+				}
+			}
+		}
+
 		if (json || ucl_reply || compact) {
 			unsigned char *json_header;
 			/* We also append json data as a specific header */
@@ -2102,8 +2143,29 @@ rspamc_client_cb(struct rspamd_client_connection *conn,
 				}
 
 				if (body) {
-					rspamc_print(out, "\nNew body:\n{}\n",
-								 std::string_view{body, bodylen});
+					if (output_body) {
+						/* Save body to file */
+						FILE *body_file = fopen(output_body, "wb");
+						if (body_file) {
+							if (fwrite(body, 1, bodylen, body_file) != bodylen) {
+								rspamc_print(stderr, "Failed to write body to file {}: {}\n",
+											 output_body, strerror(errno));
+							}
+							else {
+								rspamc_print(out, "\nNew body saved to: {}\n", output_body);
+							}
+							fclose(body_file);
+						}
+						else {
+							rspamc_print(stderr, "Failed to open output file {}: {}\n",
+										 output_body, strerror(errno));
+						}
+					}
+					else {
+						/* Print to stdout */
+						rspamc_print(out, "\nNew body:\n{}\n",
+									 std::string_view{body, bodylen});
+					}
 				}
 
 				ucl_object_unref(result);
